import tqdm, tqdm.auto, random, numpy as np, h5py
from joblib import Parallel, delayed
from multiprocessing import cpu_count
from fastdtw import fastdtw
from scipy.spatial.distance import euclidean
from sklearn.metrics import confusion_matrix
from ...internals import _Validator

class KNNClassifier:
    """A k-Nearest Neighbor classifier that compares differing length observation sequences using the efficient FastDTW dynamic time warping algorithm.

    Parameters
    ----------
    k: int
        Number of neighbors.

    radius: int
        Radius parameter for FastDTW.

        **See**: `Stan Salvador, and Philip Chan. "FastDTW: Toward accurate dynamic time warping in linear time and space." Intelligent Data Analysis 11.5 (2007), 561-580. <https://pdfs.semanticscholar.org/05a2/0cde15e172fc82f32774dd0cf4fe5827cad2.pdf>`_

    metric: callable
        Distance metric for FastDTW.

    weighting: callable
        A function that specifies how distance weighting should be performed. Using a constant-valued function to set all weights equally is equivalent to no weighting (which is the default configuration).
        Common weighting functions are :math:`e^{- \\alpha x}` or :math:`\\frac{1}{x}`, where :math:`x` is the DTW distance between two observation sequences.

        A weighting function should *ideally* be defined at :math:`x=0` in the rare event that two observation sequences are perfectly aligned
        (i.e. have zero DTW distance).

        - **Input**: :math:`x \geq 0`, a DTW distance between two observation sequences.
        - **Output**: A floating point value representing the weight used to perform nearest neighbor classification.

        .. note::
            Depending on your distance *metric*, it may be desirable to restrict DTW distances to a small range if you intend to use a weighting function.

            Using the :class:`~MinMaxScale` or :class:`~Standardize` preprocessing transformations to scale your features helps to ensure that distances remain small.
    """

    def __init__(self, k, radius, metric=euclidean, weighting=(lambda x: 1)):
        self._val = _Validator()
        self._k = self._val.restricted_integer(
            k, lambda x: x > 0, desc='number of neighbors', expected='greater than zero')
        self._radius = self._val.restricted_integer(
            radius, lambda x: x > 0, desc='radius parameter', expected='greater than zero')
        self._metric = self._val.func(metric, 'DTW distance metric')
        self._weighting = self._val.func(weighting, 'distance weighting function')

    def fit(self, X, y):
        """Fits the classifier by adding labeled training observation sequences.

        Parameters
        ----------
        X: List[numpy.ndarray]
            A list of multiple observation sequences.

        y: List[str]
            A list of labels for the observation sequences.
        """
        self._X, self._y = self._val.observation_sequences_and_labels(X, y)

    def predict(self, X, verbose=True, n_jobs=1):
        """Predicts the label for an observation sequence (or multiple sequences).

        Parameters
        ----------
        X: numpy.ndarray or List[numpy.ndarray]
            An individual observation sequence or a list of multiple observation sequences.

        verbose: bool
            Whether to display a progress bar or not.

        n_jobs: int
            | The number of jobs to run in parallel.
            | Setting this to -1 will use all available CPU cores.

        Returns
        -------
        prediction(s): str or List[str]
            The predicted label(s) for the observation sequence(s).
        """
        try:
            (self._X, self._y)
        except AttributeError:
            raise RuntimeError('The classifier needs to be fitted before predictions are made')

        X = self._val.observation_sequences(X, allow_single=True)
        self._val.boolean(verbose, desc='verbose')
        self._val.restricted_integer(n_jobs, lambda x: x == -1 or x > 0, 'number of jobs', '-1 or greater than zero')

        # FastDTW distance measure
        distance = lambda x1, x2: fastdtw(x1, x2, radius=self._radius, dist=self._metric)[0]

        if isinstance(X, np.ndarray):
            distances = [distance(X, x) for x in tqdm.auto.tqdm(self._X, desc='Calculating distances', disable=not(verbose))]
            return self._find_nearest(distances)
        else:
            if n_jobs == 1:
                labels = []
                for sequence in tqdm.auto.tqdm(X, desc='Classifying examples', disable=not(verbose)):
                    distances = [distance(sequence, x) for x in self._X]
                    labels.append(self._find_nearest(distances))
                return labels
            else:
                n_jobs = cpu_count() if n_jobs == -1 else n_jobs
                X_chunks = [list(chunk) for chunk in np.array_split(X, n_jobs)]
                labels = Parallel(n_jobs=n_jobs)(delayed(self._parallel_predict)(i+1, chunk, distance, verbose) for i, chunk in enumerate(X_chunks))
                return [label for sublist in labels for label in sublist] # Flatten the resulting array

    def _find_nearest(self, distances):
        # Find the indices of the k nearest points
        idx = np.argpartition(distances, self._k)[:self._k]

        # Calculate class scores by accumulating weighted distances
        class_scores = {}
        for i in idx:
            label = self._y[i]
            if label in class_scores:
                class_scores[label] += self._weighting(distances[i])
            else:
                class_scores[label] = 0

        # Find the labels with the maximum class score
        max_score = max(class_scores.values())
        max_labels = [k for k, v in class_scores.items() if v == max_score]

        # Randomly pick from the set of labels with the maximum class score
        return random.choice(max_labels)

    def _parallel_predict(self, process, chunk, distance, verbose):
        labels = []
        for sequence in tqdm.tqdm(chunk, desc='Classifying examples (process {})'.format(process), disable=not(verbose), position=process-1):
            distances = [distance(sequence, x) for x in self._X]
            labels.append(self._find_nearest(distances))
        return labels

    def evaluate(self, X, y, labels=None, verbose=True, n_jobs=1):
        """Evaluates the performance of the classifier on a batch of observation sequences and their labels.

        Parameters
        ----------
        X: List[numpy.ndarray]
            A list of multiple observation sequences.

        y: List[str]
            A list of labels for the observation sequences.

        labels: List[str]
            A list of labels for ordering the axes of the confusion matrix.

        verbose: bool
            Whether to display a progress bar for predictions or not.

        n_jobs: int
            | The number of jobs to run in parallel.
            | Setting this to -1 will use all available CPU cores.

        Returns
        -------
        accuracy: float
            The categorical accuracy of the classifier on the observation sequences.

        confusion: numpy.ndarray
            The confusion matrix representing the discrepancy between predicted and actual labels.
        """
        X, y = self._val.observation_sequences_and_labels(X, y)
        self._val.boolean(verbose, desc='verbose')

        if labels is not None:
            self._val.list_of_strings(labels, desc='confusion matrix labels')

        # Classify each observation sequence and calculate confusion matrix
        predictions = self.predict(X, verbose=verbose, n_jobs=n_jobs)
        cm = confusion_matrix(y, predictions, labels=labels)

        return np.sum(np.diag(cm)) / np.sum(cm), cm

    def save(self, path):
        """Stores the :class:`KNNClassifier` object into a `HDF5 <https://support.hdfgroup.org/HDF5/doc/H5.intro.html>`_ file.

        .. note:
            As :math:`k`-NN is a non-parametric classification algorithms, saving the classifier simply saves
            all of the training observation sequences and labels (along with the hyper-parameters).

        Parameters
        ----------
        path: str
            File path (with or without `.h5` extension) to store the HDF5-serialized :class:`KNNClassifier` object.
        """

        try:
            (self._X, self._y)
        except AttributeError:
            raise RuntimeError('The classifier needs to be fitted before it can be saved')

        with h5py.File(path, 'w') as f:
            # Store hyper-parameters (k, radius)
            params = f.create_group('params')
            params.create_dataset('k', data=self._k)
            params.create_dataset('radius', data=self._radius)

            # Store training data and labels (X, y)
            data = f.create_group('data')
            X = data.create_group('X')
            for i, x in enumerate(self._X):
                X.create_dataset(str(i), data=x)
            data.create_dataset('y', data=np.string_(self._y))

    @classmethod
    def load(cls, path, encoding='utf-8', metric=euclidean, weighting=(lambda x: 1)):
        """Deserializes a HDF5-serialized :class:`KNNClassifier` object.

        Parameters
        ----------
        path: str
            File path of the serialized HDF5 data generated by the :meth:`save` method.

        encoding: str
            The encoding used to represent training labels when decoding the HDF5 file.

            .. note::
                Supported string encodings in Python can be found `here <https://docs.python.org/3/library/codecs.html#standard-encodings>`_.

        metric: callable
            Distance metric for FastDTW (see :class:`KNNClassifier`).

        weighting: callable
            A function that specifies how distance weighting should be performed (see :class:`KNNClassifier`).

        Returns
        -------
        deserialized: :class:`KNNClassifier`
            The deserialized DTW :math:`k`-NN classifier object.

        See Also
        --------
        save: Serializes a :class:`KNNClassifier` into a HDF5 file.
        """

        with h5py.File(path, 'r') as f:
            # Deserialize the model hyper-parameters
            params = f['params']
            clf = cls(k=int(params['k'][()]), radius=int(params['radius'][()]), metric=metric, weighting=weighting)

            # Deserialize the training data and labels
            X, y = f['data']['X'], f['data']['y']
            clf._X = [np.array(X[k]) for k in sorted(X.keys(), key=lambda k: int(k))]
            clf._y = [label.decode(encoding) for label in y]

        return clf


import numpy as np, pomegranate as pg, json
from .topologies.ergodic import _ErgodicTopology
from .topologies.left_right import _LeftRightTopology
from .topologies.strict_left_right import _StrictLeftRightTopology
from ...internals import _Validator

class HMM:
    """A hidden Markov model representing an isolated temporal sequence class.

    Parameters
    ----------
    label: str
        A label for the model, corresponding to the class being represented.

    n_states: int
        The number of states for the model.

    topology: {'ergodic', 'left-right', 'strict-left-right'}
        The topology for the model.

    random_state: numpy.random.RandomState, int, optional
        A random state object or seed for reproducible randomness.

    Attributes
    ----------
    label: str
        The label for the model.

    n_states: int
        The number of states for the model.

    n_seqs: int
        The number of observation sequences use to train the model.

    initial: numpy.ndarray
        The initial state distribution of the model.

    transitions: numpy.ndarray
        The transition matrix of the model.
    """

    def __init__(self, label, n_states, topology='left-right', random_state=None):
        self._val = _Validator()
        self._label = self._val.string(label, 'model label')
        self._n_states = self._val.restricted_integer(
            n_states, lambda x: x > 0, desc='number of states', expected='greater than zero')
        self._val.one_of(topology, ['ergodic', 'left-right', 'strict-left-right'], desc='topology')
        self._random_state = self._val.random_state(random_state)
        self._topologies = {'ergodic': _ErgodicTopology, 'left-right': _LeftRightTopology, 'strict-left-right': _StrictLeftRightTopology}
        self._topologies.update(dict([reversed(i) for i in self._topologies.items()]))
        self._topology = self._topologies[topology](self._n_states, self._random_state)

    def set_uniform_initial(self):
        """Sets a uniform initial state distribution."""
        self._initial = self._topology.uniform_initial()

    def set_random_initial(self):
        """Sets a random initial state distribution."""
        self._initial = self._topology.random_initial()

    def set_uniform_transitions(self):
        """Sets a uniform transition matrix according to the topology."""
        self._transitions = self._topology.uniform_transitions()

    def set_random_transitions(self):
        """Sets a random transition matrix according to the topology."""
        self._transitions = self._topology.random_transitions()

    def fit(self, X, n_jobs=1):
        """Fits the HMM to observation sequences assumed to be labeled as the class that the model represents.

        Parameters
        ----------
        X: List[numpy.ndarray]
            Collection of multivariate observation sequences, each of shape :math:`(T \\times D)` where
            :math:`T` may vary per observation sequence.

        n_jobs: int
            | The number of jobs to run in parallel.
            | Setting this to -1 will use all available CPU cores.
        """
        X = self._val.observation_sequences(X)
        self._val.restricted_integer(n_jobs, lambda x: x == -1 or x > 0, 'number of jobs', '-1 or greater than zero')

        try:
            (self._initial, self._transitions)
        except AttributeError as e:
            raise AttributeError('Must specify initial state distribution and transitions before the HMM can be fitted') from e

        self._n_seqs = len(X)
        self._n_features = X[0].shape[1]

        # Create a multivariate Gaussian emission distribution using combined samples for initial parameter esimation
        dist = pg.MultivariateGaussianDistribution.from_samples(np.concatenate(X))

        # Create the HMM object
        self._model = pg.HiddenMarkovModel.from_matrix(
            name=self._label,
            transition_probabilities=self._transitions,
            distributions=[dist.copy() for _ in range(self._n_states)],
            starts=self._initial
        )

        # Perform the Baum-Welch algorithm to fit the model to the observations
        self._model.fit(X, n_jobs=n_jobs)

        # Update the initial state distribution and transitions to reflect the updated parameters
        inner_tx = self._model.dense_transition_matrix()[:, :self._n_states]
        self._initial = inner_tx[self._n_states]
        self._transitions = inner_tx[:self._n_states]

    def forward(self, sequence):
        """Runs the forward algorithm to calculate the (negative log) likelihood of the model generating an observation sequence.

        Parameters
        ----------
        sequence: numpy.ndarray
            An individual sequence of observations of size :math:`(T \\times D)` where
            :math:`T` is the number of time frames (or observations) and
            :math:`D` is the number of features.

        Returns
        -------
        negative log-likelihood: float
            The negative log-likelihood of the model generating the observation sequence.
        """
        try:
            self._model
        except AttributeError as e:
            raise AttributeError('The model must be fitted before running the forward algorithm') from e

        sequence = self._val.observation_sequences(sequence, allow_single=True)
        if not sequence.shape[1] == self._n_features:
            raise ValueError('Number of observation features must match the dimensionality of the original data used to fit the model')

        return -self._model.log_probability(sequence)

    @property
    def label(self):
        return self._label

    @property
    def n_states(self):
        return self._n_states

    @property
    def n_seqs(self):
        try:
            return self._n_seqs
        except AttributeError as e:
            raise AttributeError('The model has not been fitted and has not seen any observation sequences') from e

    @property
    def initial(self):
        try:
            return self._initial
        except AttributeError as e:
            raise AttributeError('No initial state distribution has been defined') from e

    @initial.setter
    def initial(self, probabilities):
        self._topology.validate_initial(probabilities)
        self._initial = probabilities

    @property
    def transitions(self):
        try:
            return self._transitions
        except AttributeError as e:
            raise AttributeError('No transition matrix has been defined') from e

    @transitions.setter
    def transitions(self, probabilities):
        self._topology.validate_transitions(probabilities)
        self._transitions = probabilities

    def as_dict(self):
        """Serializes the :class:`HMM` object into a `dict`, ready to be stored in JSON format.

        Returns
        -------
        serialized: dict
            JSON-ready serialization of the :class:`HMM` object.
        """

        try:
            self._model
        except AttributeError as e:
            raise AttributeError('The model needs to be fitted before it can be exported to a dict') from e

        model = self._model.to_json()

        if 'NaN' in model:
            raise ValueError('Encountered NaN value(s) in HMM parameters')
        else:
            return {
                'type': 'HMM',
                'label': self._label,
                'n_states': self._n_states,
                'topology': self._topologies[self._topology.__class__],
                'model': {
                    'initial': self._initial.tolist(),
                    'transitions': self._transitions.tolist(),
                    'n_seqs': self._n_seqs,
                    'n_features': self._n_features,
                    'hmm': json.loads(model)
                }
            }

    def save(self, path):
        """Converts the :class:`HMM` object into a `dict` and stores it in a JSON file.

        Parameters
        ----------
        path: str
            File path (with or without `.json` extension) to store the JSON-serialized :class:`HMM` object.

        See Also
        --------
        as_dict: Generates the `dict` that is stored in the JSON file.
        """

        data = self.as_dict()
        with open(path, 'w') as f:
            json.dump(data, f, indent=4)

    @classmethod
    def load(cls, data, random_state=None):
        """Deserializes either a `dict` or JSON serialized :class:`HMM` object.

        Parameters
        ----------
        data: str or dict
            - File path of the serialized JSON data generated by the :meth:`save` method.
            - `dict` representation of the :class:`HMM`, generated by the :meth:`as_dict` method.

        random_state: numpy.random.RandomState, int, optional
            A random state object or seed for reproducible randomness.

        Returns
        -------
        deserialized: :class:`HMM`
            The deserialized HMM object.

        See Also
        --------
        save: Serializes a :class:`HMM` into a JSON file.
        as_dict: Generates a `dict` representation of the :class:`HMM`.
        """

        # Load the serialized HMM data
        if isinstance(data, dict):
            pass
        elif isinstance(data, str):
            with open(data, 'r') as f:
                data = json.load(f)
        else:
            pass

        # Check that JSON is in the "correct" format
        if data['type'] == 'HMM':
            pass
        elif data['type'] == 'GMMHMM':
            raise ValueError('You must use the GMMHMM class to deserialize a stored GMMHMM model')
        else:
            raise ValueError("Attempted to deserialize an invalid model - expected 'type' field to be 'HMM'")

        # Deserialize the data into a HMM object
        hmm = cls(data['label'], data['n_states'], data['topology'], random_state=random_state)
        hmm._initial = np.array(data['model']['initial'])
        hmm._transitions = np.array(data['model']['transitions'])
        hmm._n_seqs = data['model']['n_seqs']
        hmm._n_features = data['model']['n_features']
        hmm._model = pg.HiddenMarkovModel.from_json(json.dumps(data['model']['hmm']))

        return hmm


import pytest, warnings, os, json, numpy as np
from copy import deepcopy
with warnings.catch_warnings():
    warnings.filterwarnings('ignore', category=DeprecationWarning)
    import pomegranate as pg
from sequentia.classifiers import HMM, GMMHMM, _LeftRightTopology, _ErgodicTopology, _StrictLeftRightTopology
from ....support import assert_equal, assert_not_equal

# Set seed for reproducible randomness
seed = 0
np.random.seed(seed)
rng = np.random.RandomState(seed)

# Create some sample data
X = [rng.random((10 * i, 3)) for i in range(1, 4)]
x = rng.random((15, 3))

# Unparameterized HMMs
hmm_diag = GMMHMM(label='c1', n_states=5, n_components=5, covariance='diagonal', random_state=rng)
hmm_full = GMMHMM(label='c1', n_states=5, n_components=5, covariance='full', random_state=rng)

# ============================== #
# GMMHMM.n_components (property) #
# ============================== #

def test_n_components():
    assert deepcopy(hmm_diag).n_components == 5

# ============================ #
# GMMHMM.covariance (property) #
# ============================ #

def test_covariance():
    assert deepcopy(hmm_diag).covariance == 'diagonal'

# ================ #
# GMMHMM.as_dict() #
# ================ #

def test_as_dict_unfitted():
    """Export an unfitted GMMHMM to dict"""
    hmm = deepcopy(hmm_diag)
    with pytest.raises(AttributeError) as e:
        hmm.as_dict()
    assert str(e.value) == 'The model needs to be fitted before it can be exported to a dict'

def test_as_dict_fitted():
    """Export a fitted GMMHMM to dict"""
    hmm = deepcopy(hmm_diag)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    before = hmm.initial, hmm.transitions
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        hmm.fit(X)
    d = hmm.as_dict()

    assert d['type'] == 'GMMHMM'
    assert d['label'] == 'c1'
    assert d['n_states'] == 5
    assert d['n_components'] == 5
    assert d['covariance'] == 'diagonal'
    assert d['topology'] == 'left-right'
    assert_not_equal(d['model']['initial'], before[0])
    assert_not_equal(d['model']['transitions'], before[1])
    assert d['model']['n_seqs'] == 3
    assert d['model']['n_features'] == 3
    assert isinstance(d['model']['hmm'], dict)

# ========== #
# HMM.save() #
# ========== #

def test_save_directory():
    """Save a GMMHMM into a directory"""
    hmm = deepcopy(hmm_diag)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        hmm.fit(X)
    try:
        with pytest.raises(IsADirectoryError) as e:
            hmm.save('.')
        assert str(e.value) == "[Errno 21] Is a directory: '.'"
    except ValueError:
        pass

def test_save_no_extension():
    """Save a GMMHMM into a file without an extension"""
    try:
        hmm = deepcopy(hmm_diag)
        hmm.set_uniform_initial()
        hmm.set_uniform_transitions()
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=DeprecationWarning)
            hmm.fit(X)
        try:
            hmm.save('test')
            assert os.path.isfile('test')
        except ValueError:
            pass
    finally:
        try:
            os.remove('test')
        except OSError:
            pass

def test_save_with_extension():
    """Save a GMMHMM into a file with a .json extension"""
    try:
        hmm = deepcopy(hmm_diag)
        hmm.set_uniform_initial()
        hmm.set_uniform_transitions()
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=DeprecationWarning)
            hmm.fit(X)
        try:
            hmm.save('test.json')
            assert os.path.isfile('test.json')
        except ValueError:
            pass
    finally:
        try:
            os.remove('test.json')
        except OSError:
            pass

# ============= #
# GMMHMM.load() #
# ============= #

def test_load_invalid_dict():
    """Load a GMMHMM from an invalid dict"""
    with pytest.raises(KeyError) as e:
        GMMHMM.load({})

def test_load_dict():
    """Load a GMMHMM from a valid dict"""
    hmm = deepcopy(hmm_diag)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    before = hmm.initial, hmm.transitions
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        hmm.fit(X)
    hmm = GMMHMM.load(hmm.as_dict())

    assert isinstance(hmm, GMMHMM)
    assert hmm._label == 'c1'
    assert hmm._n_states == 5
    assert hmm._n_components == 5
    assert hmm._covariance == 'diagonal'
    assert isinstance(hmm._topology, _LeftRightTopology)
    assert_not_equal(hmm._initial, before[0])
    assert_not_equal(hmm._transitions, before[1])
    assert hmm._n_seqs == 3
    assert hmm._n_features == 3
    assert isinstance(hmm._model, pg.HiddenMarkovModel)

def test_load_invalid_path():
    """Load a GMMHMM from a directory"""
    with pytest.raises(IsADirectoryError) as e:
        GMMHMM.load('.')

def test_load_inexistent_path():
    """Load a GMMHMM from an inexistent path"""
    with pytest.raises(FileNotFoundError) as e:
        GMMHMM.load('test')

def test_load_invalid_format():
    """Load a GMMHMM from an illegally formatted file"""
    try:
        with open('test', 'w') as f:
            f.write('illegal')
        with pytest.raises(json.decoder.JSONDecodeError) as e:
            GMMHMM.load('test')
    finally:
        os.remove('test')

def test_load_invalid_json():
    """Load a GMMHMM from an invalid JSON file"""
    try:
        with open('test', 'w') as f:
            f.write("{}")
        with pytest.raises(KeyError) as e:
            GMMHMM.load('test')
    finally:
        os.remove('test')

def test_load_path():
    """Load a GMMHMM from a valid JSON file"""
    try:
        hmm = deepcopy(hmm_diag)
        hmm.set_uniform_initial()
        hmm.set_uniform_transitions()
        before = hmm.initial, hmm.transitions
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=DeprecationWarning)
            hmm.fit(X)
        hmm.save('test')
        hmm = GMMHMM.load('test')

        assert isinstance(hmm, GMMHMM)
        assert hmm._label == 'c1'
        assert hmm._n_states == 5
        assert hmm._n_components == 5
        assert isinstance(hmm._topology, _LeftRightTopology)
        assert hmm._covariance == 'diagonal'
        assert_not_equal(hmm._initial, before[0])
        assert_not_equal(hmm._transitions, before[1])
        assert hmm._n_seqs == 3
        assert hmm._n_features == 3
        assert isinstance(hmm._model, pg.HiddenMarkovModel)
    finally:
        os.remove('test')


import pytest, warnings, os, json, numpy as np
from copy import deepcopy
with warnings.catch_warnings():
    warnings.filterwarnings('ignore', category=DeprecationWarning)
    import pomegranate as pg
from sequentia.classifiers import HMM, _LeftRightTopology, _ErgodicTopology, _StrictLeftRightTopology
from ....support import assert_equal, assert_not_equal

# Set seed for reproducible randomness
seed = 0
np.random.seed(seed)
rng = np.random.RandomState(seed)

# Create some sample data
X = [rng.random((10 * i, 3)) for i in range(1, 4)]
x = rng.random((15, 3))

# Unparameterized HMMs
hmm_lr = HMM(label='c1', n_states=5, topology='left-right', random_state=rng)
hmm_e = HMM(label='c1', n_states=5, topology='ergodic', random_state=rng)
hmm_slr = HMM(label='c1', n_states=5, topology='strict-left-right', random_state=rng)

# ================================================== #
# HMM.set_uniform_initial() + HMM.initial (property) #
# ================================================== #

def test_left_right_uniform_initial():
    """Uniform initial state distribution for a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_uniform_initial()
    assert_equal(hmm.initial, np.array([
        0.2, 0.2, 0.2, 0.2, 0.2
    ]))

def test_ergodic_uniform_initial():
    """Uniform initial state distribution for an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_uniform_initial()
    assert_equal(hmm.initial, np.array([
        0.2, 0.2, 0.2, 0.2, 0.2
    ]))

def test_strict_left_right_uniform_initial():
    """Uniform initial state distribution for a strict left-right HMM"""
    hmm = deepcopy(hmm_slr)
    hmm.set_uniform_initial()
    assert_equal(hmm.initial, np.array([
        0.2, 0.2, 0.2, 0.2, 0.2
    ]))

# ================================================= #
# HMM.set_random_initial() + HMM.initial (property) #
# ================================================= #

def test_left_right_random_initial():
    """Random initial state distribution for a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    assert_equal(hmm.initial, np.array([
        0.35029635, 0.13344569, 0.02784745, 0.33782453, 0.15058597
    ]))

def test_ergodic_random_initial():
    """Random initial state distribution for an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_random_initial()
    assert_equal(hmm.initial, np.array([
        0.35029635, 0.13344569, 0.02784745, 0.33782453, 0.15058597
    ]))

def test_strict_left_right_random_initial():
    """Random initial state distribution for a strict left-right HMM"""
    hmm = deepcopy(hmm_slr)
    hmm.set_random_initial()
    assert_equal(hmm.initial, np.array([
        0.35029635, 0.13344569, 0.02784745, 0.33782453, 0.15058597
    ]))

# ========================================================== #
# HMM.set_uniform_transitions() + HMM.transitions (property) #
# ========================================================== #

def test_left_right_uniform_transitions():
    """Uniform transition matrix for a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_uniform_transitions()
    assert_equal(hmm.transitions, np.array([
        [0.2, 0.2 , 0.2       , 0.2       , 0.2       ],
        [0. , 0.25, 0.25      , 0.25      , 0.25      ],
        [0. , 0.  , 0.33333333, 0.33333333, 0.33333333],
        [0. , 0.  , 0.        , 0.5       , 0.5       ],
        [0. , 0.  , 0.        , 0.        , 1.        ]
    ]))

def test_ergodic_uniform_transitions():
    """Uniform transition matrix for an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_uniform_transitions()
    assert_equal(hmm.transitions, np.array([
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2]
    ]))

def test_strict_left_right_uniform_transitions():
    """Uniform transition matrix for a strict left-right HMM"""
    hmm = deepcopy(hmm_slr)
    hmm.set_uniform_transitions()
    assert_equal(hmm.transitions, np.array([
        [0.5, 0.5, 0. , 0. , 0. ],
        [0. , 0.5, 0.5, 0. , 0. ],
        [0. , 0. , 0.5, 0.5, 0. ],
        [0. , 0. , 0. , 0.5, 0.5],
        [0. , 0. , 0. , 0. , 1. ]
    ]))

# ========================================================= #
# HMM.set_random_transitions() + HMM.transitions (property) #
# ========================================================= #

def test_left_right_random_transitions():
    """Random transition matrix for a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_transitions()
    assert_equal(hmm.transitions, np.array([
        [0.35029635, 0.13344569, 0.02784745, 0.33782453, 0.15058597],
        [0.        , 0.20580715, 0.5280311 , 0.06521685, 0.20094491],
        [0.        , 0.        , 0.33761567, 0.2333124 , 0.42907193],
        [0.        , 0.        , 0.        , 0.36824778, 0.63175222],
        [0.        , 0.        , 0.        , 0.        , 1.        ]
    ]))

def test_ergodic_random_transitions():
    """Random transition matrix for an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_random_transitions()
    assert_equal(hmm.transitions, np.array([
        [0.35029635, 0.13344569, 0.02784745, 0.33782453, 0.15058597],
        [0.19252534, 0.15767581, 0.47989976, 0.01708551, 0.15281357],
        [0.19375092, 0.16425506, 0.21828034, 0.11397708, 0.30973661],
        [0.46906977, 0.02941216, 0.17137502, 0.0333193 , 0.29682374],
        [0.21312406, 0.35221103, 0.08556524, 0.06613143, 0.28296824]
    ]))

def test_strict_left_right_random_transitions():
    """Random transition matrix for a strict left-right HMM"""
    hmm = deepcopy(hmm_slr)
    hmm.set_random_transitions()
    assert_equal(hmm.transitions, np.array([
        [0.72413873, 0.27586127, 0.        , 0.        , 0.        ],
        [0.        , 0.07615418, 0.92384582, 0.        , 0.        ],
        [0.        , 0.        , 0.81752797, 0.18247203, 0.        ],
        [0.        , 0.        , 0.        , 0.24730529, 0.75269471],
        [0.        , 0.        , 0.        , 0.        , 1.        ]
    ]))

# ====================== #
# HMM.fit() + HMM.n_seqs #
# ====================== #

def test_fit_without_initial_and_transition():
    """Fitting before setting the initial state distribution and transition matrix"""
    hmm = deepcopy(hmm_lr)
    with warnings.catch_warnings(), pytest.raises(AttributeError) as e:
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        hmm.fit(X)
    assert str(e.value) == 'Must specify initial state distribution and transitions before the HMM can be fitted'

def test_fit_without_transitions():
    """Fitting before setting the initial transition matrix"""
    hmm = deepcopy(hmm_lr)
    hmm.set_uniform_initial()
    with warnings.catch_warnings(), pytest.raises(AttributeError) as e:
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        hmm.fit(X)
    assert str(e.value) == 'Must specify initial state distribution and transitions before the HMM can be fitted'

def test_fit_without_initial():
    """Fitting before setting the initial state distribution"""
    hmm = deepcopy(hmm_lr)
    hmm.set_uniform_transitions()
    with warnings.catch_warnings(), pytest.raises(AttributeError) as e:
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        hmm.fit(X)
    assert str(e.value) == 'Must specify initial state distribution and transitions before the HMM can be fitted'

def test_fit_sets_internals():
    """Check that fitting sets internal attributes"""
    hmm = deepcopy(hmm_lr)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        hmm.fit(X)
    assert hmm.n_seqs == 3
    assert isinstance(hmm._model, pg.HiddenMarkovModel)

def test_left_right_fit_updates_initial():
    """Check that fitting updates the initial state distribution of a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    assert_equal(hmm.initial, np.array([
        0.2, 0.2, 0.2, 0.2, 0.2
    ]))
    before = hmm.initial
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        hmm.fit(X)
    assert_not_equal(hmm.initial, before)

def test_left_right_fit_updates_transitions():
    """Check that fitting updates the transition matrix of a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    before = hmm.transitions
    assert_equal(hmm.transitions, np.array([
        [0.2       , 0.2       , 0.2       , 0.2       , 0.2       ],
        [0.        , 0.25      , 0.25      , 0.25      , 0.25      ],
        [0.        , 0.        , 0.33333333, 0.33333333, 0.33333333],
        [0.        , 0.        , 0.        , 0.5       , 0.5       ],
        [0.        , 0.        , 0.        , 0.        , 1.        ]
    ]))
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        hmm.fit(X)
    assert_not_equal(hmm.transitions, before)

def test_ergodic_fit_doesnt_updates_initial():
    """Check that fitting does not update the initial state distribution of an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    assert_equal(hmm.initial, np.array([
        0.2, 0.2, 0.2, 0.2, 0.2
    ]))
    before = hmm.initial
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        hmm.fit(X)
    assert_equal(hmm.initial, before)

def test_ergodic_fit_doesnt_updates_transitions():
    """Check that fitting does not update the transition matrix of an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    assert_equal(hmm.transitions, np.array([
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2]
    ]))
    before = hmm.transitions
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        hmm.fit(X)
    assert_equal(hmm.transitions, before)

def test_strict_left_right_fit_updates_initial():
    """Check that fitting updates the initial state distribution of a strict left-right HMM"""
    hmm = deepcopy(hmm_slr)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    assert_equal(hmm.initial, np.array([
        0.2, 0.2, 0.2, 0.2, 0.2
    ]))
    before = hmm.initial
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        hmm.fit(X)
    assert_not_equal(hmm.initial, before)

def test_strict_left_right_fit_updates_transitions():
    """Check that fitting updates the transition matrix of a strict left-right HMM"""
    hmm = deepcopy(hmm_slr)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    before = hmm.transitions
    assert_equal(hmm.transitions, np.array([
        [0.5       , 0.5       , 0.        , 0.        , 0.        ],
        [0.        , 0.5       , 0.5       , 0.        , 0.        ],
        [0.        , 0.        , 0.5       , 0.5       , 0.        ],
        [0.        , 0.        , 0.        , 0.5       , 0.5       ],
        [0.        , 0.        , 0.        , 0.        , 1.        ]
    ]))
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        hmm.fit(X)
    assert_not_equal(hmm.transitions, before)

# ============= #
# HMM.forward() #
# ============= #

def test_forward_without_fit():
    """Forward algorithm without fitting the HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    with pytest.raises(AttributeError) as e:
        hmm.forward(x)
    assert str(e.value) == 'The model must be fitted before running the forward algorithm'

def test_left_right_forward():
    """Forward algorithm on a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        hmm.fit(X)
    assert isinstance(hmm.forward(x), float)

def test_ergodic_forward():
    """Forward algorithm on an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        hmm.fit(X)
    assert isinstance(hmm.forward(x), float)

def test_strict_left_right_forward():
    """Forward algorithm on a strict left-right HMM"""
    hmm = deepcopy(hmm_slr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        hmm.fit(X)
    assert isinstance(hmm.forward(x), float)

# ==================== #
# HMM.label (property) #
# ==================== #

def test_label():
    assert deepcopy(hmm_lr).label == 'c1'

# ======================= #
# HMM.n_states (property) #
# ======================= #

def test_n_states():
    assert deepcopy(hmm_lr).n_states == 5

# ===================== #
# HMM.n_seqs (property) #
# ===================== #

def test_n_seqs_without_fit():
    """Number of sequences without fitting the HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    with pytest.raises(AttributeError) as e:
        hmm.n_seqs
    assert str(e.value) == 'The model has not been fitted and has not seen any observation sequences'

def test_n_seqs_with_fit():
    """Number of sequences after fitting the HMM"""
    hmm = deepcopy(hmm_lr)
    hmm.set_random_initial()
    hmm.set_random_transitions()
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        hmm.fit(X)
    assert hmm.n_seqs == 3

# ====================== #
# HMM.initial (property) #
# ====================== #

def test_initial_without_setting():
    """Get initial state distribution without setting it"""
    hmm = deepcopy(hmm_lr)
    with pytest.raises(AttributeError) as e:
        hmm.initial
    assert str(e.value) == 'No initial state distribution has been defined'

# ========================== #
# HMM.transitions (property) #
# ========================== #

def test_transitions_without_setting():
    """Get transition matrix without setting it"""
    hmm = deepcopy(hmm_lr)
    with pytest.raises(AttributeError) as e:
        hmm.transitions
    assert str(e.value) == 'No transition matrix has been defined'

# ==================== #
# HMM.initial (setter) #
# ==================== #

def test_left_right_initial_left_right():
    """Set an initial state distribution generated by a left-right topology on a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    topology = _LeftRightTopology(n_states=5, random_state=rng)
    initial = topology.random_initial()
    hmm.initial = initial
    assert_equal(hmm.initial, initial)

def test_left_right_initial_ergodic():
    """Set an initial state distribution generated by a left-right topology on an ergodic HMM"""
    hmm = deepcopy(hmm_lr)
    topology = _ErgodicTopology(n_states=5, random_state=rng)
    initial = topology.random_initial()
    hmm.initial = initial
    assert_equal(hmm.initial, initial)

def test_left_right_initial_strict_left_right():
    """Set an initial state distribution generated by a left-right topology on an strict left-right HMM"""
    hmm = deepcopy(hmm_lr)
    topology = _StrictLeftRightTopology(n_states=5, random_state=rng)
    initial = topology.random_initial()
    hmm.initial = initial
    assert_equal(hmm.initial, initial)

def test_ergodic_initial_left_right():
    """Set an initial state distribution generated by an ergodic topology on a left-right HMM"""
    hmm = deepcopy(hmm_e)
    topology = _LeftRightTopology(n_states=5, random_state=rng)
    initial = topology.random_initial()
    hmm.initial = initial
    assert_equal(hmm.initial, initial)

def test_ergodic_initial_ergodic():
    """Set an initial state distribution generated by an ergodic topology on an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    topology = _ErgodicTopology(n_states=5, random_state=rng)
    initial = topology.random_initial()
    hmm.initial = initial
    assert_equal(hmm.initial, initial)

def test_ergodic_initial_strict_left_right():
    """Set an initial state distribution generated by an ergodic topology on a strict left-right HMM"""
    hmm = deepcopy(hmm_e)
    topology = _StrictLeftRightTopology(n_states=5, random_state=rng)
    initial = topology.random_initial()
    hmm.initial = initial
    assert_equal(hmm.initial, initial)

def test_strict_left_right_initial_left_right():
    """Set an initial state distribution generated by a strict left-right topology on a left-right HMM"""
    hmm = deepcopy(hmm_slr)
    topology = _LeftRightTopology(n_states=5, random_state=rng)
    initial = topology.random_initial()
    hmm.initial = initial
    assert_equal(hmm.initial, initial)

def test_strict_left_right_initial_ergodic():
    """Set an initial state distribution generated by a strict left-right topology on an ergodic HMM"""
    hmm = deepcopy(hmm_slr)
    topology = _ErgodicTopology(n_states=5, random_state=rng)
    initial = topology.random_initial()
    hmm.initial = initial
    assert_equal(hmm.initial, initial)

def test_strict_left_right_initial_strict_left_right():
    """Set an initial state distribution generated by a strict left-right topology on an strict left-right HMM"""
    hmm = deepcopy(hmm_slr)
    topology = _StrictLeftRightTopology(n_states=5, random_state=rng)
    initial = topology.random_initial()
    hmm.initial = initial
    assert_equal(hmm.initial, initial)

# ======================== #
# HMM.transitions (setter) #
# ======================== #

def test_left_right_transitions_left_right():
    """Set a transition matrix generated by a left-right topology on a left-right HMM"""
    hmm = deepcopy(hmm_lr)
    topology = _LeftRightTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    hmm.transitions = transitions
    assert_equal(hmm.transitions, transitions)

def test_left_right_transitions_ergodic():
    """Set a transition matrix generated by a left-right topology on an ergodic HMM"""
    hmm = deepcopy(hmm_lr)
    topology = _ErgodicTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    with pytest.raises(ValueError) as e:
        hmm.transitions = transitions
    assert str(e.value) == 'Left-right transition matrix must be upper-triangular'

def test_left_right_transitions_strict_left_right():
    """Set a transition matrix generated by a left-right topology on a strict left-right HMM"""
    hmm = deepcopy(hmm_lr)
    topology = _StrictLeftRightTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    hmm.transitions = transitions
    assert_equal(hmm.transitions, transitions)

def test_ergodic_transitions_left_right():
    """Set a transition matrix generated by an ergodic topology on a left-right HMM"""
    hmm = deepcopy(hmm_e)
    topology = _LeftRightTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    with pytest.warns(UserWarning) as w:
        hmm.transitions = transitions
    assert w[0].message.args[0] == 'Zero probabilities in ergodic transition matrix - these transition probabilities will not be learned'
    assert_equal(hmm.transitions, transitions)

def test_ergodic_transitions_ergodic():
    """Set a transition matrix generated by an ergodic topology on an ergodic HMM"""
    hmm = deepcopy(hmm_e)
    topology = _ErgodicTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    hmm.transitions = transitions
    assert_equal(hmm.transitions, transitions)

def test_ergodic_transitions_strict_left_right():
    """Set a transition matrix generated by an ergodic topology on a strict left-right HMM"""
    hmm = deepcopy(hmm_e)
    topology = _StrictLeftRightTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    with pytest.warns(UserWarning) as w:
        hmm.transitions = transitions
    assert w[0].message.args[0] == 'Zero probabilities in ergodic transition matrix - these transition probabilities will not be learned'
    assert_equal(hmm.transitions, transitions)

def test_strict_left_right_transitions_left_right():
    """Set a transition matrix generated by a strict left-right topology on a left-right HMM"""
    hmm = deepcopy(hmm_slr)
    topology = _LeftRightTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    with pytest.raises(ValueError) as e:
        hmm.transitions = transitions
    assert str(e.value) == 'Strict left-right transition matrix must only consist of a diagonal and upper diagonal'

def test_strict_left_right_transitions_ergodic():
    """Set a transition matrix generated by a strict left-right topology on an ergodic HMM"""
    hmm = deepcopy(hmm_slr)
    topology = _ErgodicTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    with pytest.raises(ValueError) as e:
        hmm.transitions = transitions
    assert str(e.value) == 'Left-right transition matrix must be upper-triangular'

def test_strict_left_right_transitions_strict_left_right():
    """Set a transition matrix generated by a strict left-right topology on a strict left-right HMM"""
    hmm = deepcopy(hmm_slr)
    topology = _StrictLeftRightTopology(n_states=5, random_state=rng)
    transitions = topology.random_transitions()
    hmm.transitions = transitions
    assert_equal(hmm.transitions, transitions)

# ============= #
# HMM.as_dict() #
# ============= #

def test_as_dict_unfitted():
    """Export an unfitted HMM to dict"""
    hmm = deepcopy(hmm_e)
    with pytest.raises(AttributeError) as e:
        hmm.as_dict()
    assert str(e.value) == 'The model needs to be fitted before it can be exported to a dict'

def test_as_dict_fitted():
    """Export a fitted HMM to dict"""
    hmm = deepcopy(hmm_e)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        hmm.fit(X)
    d = hmm.as_dict()

    assert d['type'] == 'HMM'
    assert d['label'] == 'c1'
    assert d['n_states'] == 5
    assert d['topology'] == 'ergodic'
    assert_equal(d['model']['initial'], np.array([
        0.2, 0.2, 0.2, 0.2, 0.2
    ]))
    assert_equal(d['model']['transitions'], np.array([
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2],
        [0.2, 0.2, 0.2, 0.2, 0.2]
    ]))
    assert d['model']['n_seqs'] == 3
    assert d['model']['n_features'] == 3
    assert isinstance(d['model']['hmm'], dict)

# ========== #
# HMM.save() #
# ========== #

def test_save_directory():
    """Save a HMM into a directory"""
    hmm = deepcopy(hmm_e)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        hmm.fit(X)
    with pytest.raises(IsADirectoryError) as e:
        hmm.save('.')
    assert str(e.value) == "[Errno 21] Is a directory: '.'"

def test_save_no_extension():
    """Save a HMM into a file without an extension"""
    try:
        hmm = deepcopy(hmm_e)
        hmm.set_uniform_initial()
        hmm.set_uniform_transitions()
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=DeprecationWarning)
            hmm.fit(X)
        hmm.save('test')
        assert os.path.isfile('test')
    finally:
        os.remove('test')

def test_save_with_extension():
    """Save a HMM into a file with a .json extension"""
    try:
        hmm = deepcopy(hmm_e)
        hmm.set_uniform_initial()
        hmm.set_uniform_transitions()
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=DeprecationWarning)
            hmm.fit(X)
        hmm.save('test.json')
        assert os.path.isfile('test.json')
    finally:
        os.remove('test.json')

# ========== #
# HMM.load() #
# ========== #

def test_load_invalid_dict():
    """Load a HMM from an invalid dict"""
    with pytest.raises(KeyError) as e:
        HMM.load({})

def test_load_dict():
    """Load a HMM from a valid dict"""
    hmm = deepcopy(hmm_lr)
    hmm.set_uniform_initial()
    hmm.set_uniform_transitions()
    before = hmm.initial, hmm.transitions
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        hmm.fit(X)
    try:
        hmm = HMM.load(hmm.as_dict())

        assert isinstance(hmm, HMM)
        assert hmm._label == 'c1'
        assert hmm._n_states == 5
        assert isinstance(hmm._topology, _LeftRightTopology)
        assert_not_equal(hmm._initial, before[0])
        assert_not_equal(hmm._transitions, before[1])
        assert hmm._n_seqs == 3
        assert hmm._n_features == 3
        assert isinstance(hmm._model, pg.HiddenMarkovModel)
    except:
        pass

def test_load_invalid_path():
    """Load a HMM from a directory"""
    with pytest.raises(IsADirectoryError) as e:
        HMM.load('.')

def test_load_inexistent_path():
    """Load a HMM from an inexistent path"""
    with pytest.raises(FileNotFoundError) as e:
        HMM.load('test')

def test_load_invalid_format():
    """Load a HMM from an illegally formatted file"""
    try:
        with open('test', 'w') as f:
            f.write('illegal')
        with pytest.raises(json.decoder.JSONDecodeError) as e:
            HMM.load('test')
    finally:
        os.remove('test')

def test_load_invalid_json():
    """Load a HMM from an invalid JSON file"""
    try:
        with open('test', 'w') as f:
            f.write("{}")
        with pytest.raises(KeyError) as e:
            HMM.load('test')
    finally:
        os.remove('test')

def test_load_path():
    """Load a HMM from a valid JSON file"""
    try:
        hmm = deepcopy(hmm_slr)
        hmm.set_uniform_initial()
        hmm.set_uniform_transitions()
        before = hmm.initial, hmm.transitions
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=DeprecationWarning)
            hmm.fit(X)
        hmm.save('test')
        hmm = HMM.load('test')

        assert isinstance(hmm, HMM)
        assert hmm._label == 'c1'
        assert hmm._n_states == 5
        assert isinstance(hmm._topology, _StrictLeftRightTopology)
        assert_not_equal(hmm._initial, before[0])
        assert_not_equal(hmm._transitions, before[1])
        assert hmm._n_seqs == 3
        assert hmm._n_features == 3
        assert isinstance(hmm._model, pg.HiddenMarkovModel)
    finally:
        os.remove('test')


import numpy as np, json
from .hmm import HMM
from .gmmhmm import GMMHMM
from sklearn.metrics import confusion_matrix
from ...internals import _Validator

class HMMClassifier:
    """A classifier that combines individual :class:`~HMM` and/or :class:`~GMMHMM` objects,
    which model isolated sequences from different classes."""

    def __init__(self):
        self._val = _Validator()

    def fit(self, models):
        """Fits the classifier with a collection of :class:`~HMM` and/or :class:`~GMMHMM` objects.

        Parameters
        ----------
        models: List[HMM, GMMHMM] or Dict[Any, HMM/GMMHMM]
            A collection of :class:`~HMM` objects to use for classification.
        """
        if isinstance(models, list):
            if not all(isinstance(model, HMM) for model in models):
                raise TypeError('Expected all models to be HMM objects')
        elif isinstance(models, dict):
            values = list(models.values())
            if not all(isinstance(model, HMM) for model in values):
                raise TypeError('Expected all models to be HMM objects')
            models = values
        else:
            raise TypeError('Expected models to be a list or dict of HMM objects')

        if len(models) > 0:
            self._models = models
        else:
            raise RuntimeError('Must fit the classifier with at least one HMM')

    def predict(self, X, prior='frequency', return_scores=False):
        """Predicts the label for an observation sequence (or multiple sequences) according to maximum likelihood or posterior scores.

        Parameters
        ----------
        X: numpy.ndarray or List[numpy.ndarray]
            An individual observation sequence or a list of multiple observation sequences.

        prior: {'frequency', 'uniform'} or Dict[str, float]
            How the prior for each model is calculated to perform MAP estimation by scoring with
            the joint probability (or un-normalized posterior) :math:`\mathbb{P}(O, \lambda_c)=\mathbb{P}(O|\lambda_c)\mathbb{P}(\lambda_c)`,
            where the likelihood :math:`\mathbb{P}(O|\lambda_c)` is generated from the models' :func:`~HMM.forward` function.

            - `'frequency'`: Calculate the prior :math:`\mathbb{P}(\lambda_c)` as the proportion of training examples in class :math:`c`.
            - `'uniform'`: Set the priors uniformly such that :math:`\mathbb{P}(\lambda_c)=\\frac{1}{C}` for each class :math:`c\in\{1,\ldots,C\}`.

            Alternatively, class priors can be specified in a ``dict``, e.g. ``{'class1': 0.1, 'class2': 0.3, 'class3': 0.6}``.

        return_scores: bool
            Whether to return the scores of each model on the observation sequence(s).

        Returns
        -------
        prediction(s): str or List[str]
            The predicted label(s) for the observation sequence(s).

            If ``return_scores`` is true, then for each observation sequence, a tuple `(label, scores)` is returned for each label,
            consisting of the `scores` of each HMM and the `label` of the HMM with the best score.
        """
        X = self._val.observation_sequences(X, allow_single=True)
        if isinstance(prior, dict):
            assert len(prior) == len(self._models), 'There must be a class prior for each HMM or class'
            assert all(model.label in prior for model in self._models), 'There must be a class prior for each HMM or class'
            assert all(isinstance(p, (int, float)) for p in prior.values()), 'Class priors must be numerical'
            assert all(0. <= p <= 1. for p in prior.values()), 'Class priors must each be between zero and one'
            assert np.isclose(sum(prior.values()), 1.), 'Class priors must form a probability distribution by summing to one'
        else:
            self._val.one_of(prior, ['frequency', 'uniform'], desc='prior')
        self._val.boolean(return_scores, desc='return_scores')

        try:
            self._models
        except AttributeError as e:
            raise AttributeError('The classifier needs to be fitted before predictions are made') from e

        if prior == 'frequency':
            total_seqs = sum(model.n_seqs for model in self._models)
            prior = {model.label:(model.n_seqs / total_seqs) for model in self._models}
        elif prior == 'uniform':
            prior = {model.label:(1. / len(self._models)) for model in self._models}

        def _map(sequence):
            scores = [(model.label, model.forward(sequence) - np.log(prior[model.label])) for model in self._models]
            best = min(scores, key=lambda x: x[1])
            return (best[0], scores) if return_scores else best[0]

        # Return MAP predictions (and scores) for observation sequence(s)
        return _map(X) if isinstance(X, np.ndarray) else [_map(sequence) for sequence in X]

    def evaluate(self, X, y, prior='frequency', labels=None):
        """Evaluates the performance of the classifier on a batch of observation sequences and their labels.

        Parameters
        ----------
        X: List[numpy.ndarray]
            A list of multiple observation sequences.

        y: List[str]
            A list of labels for the observation sequences.

        prior: {'frequency', 'uniform'} or Dict[str, float]
            How the prior for each model is calculated to perform MAP estimation by scoring with
            the joint probability (or un-normalized posterior) :math:`\mathbb{P}(O, \lambda_c)=\mathbb{P}(O|\lambda_c)\mathbb{P}(\lambda_c)`,
            where the likelihood :math:`\mathbb{P}(O|\lambda_c)` is generated from the models' :func:`~HMM.forward` function.

            - `'frequency'`: Calculate the prior :math:`\mathbb{P}(\lambda_c)` as the proportion of training examples in class :math:`c`.
            - `'uniform'`: Set the priors uniformly such that :math:`\mathbb{P}(\lambda_c)=\\frac{1}{C}` for each class :math:`c\in\{1,\ldots,C\}`.

            Alternatively, class priors can be specified in a ``dict``, e.g. ``{'class1': 0.1, 'class2': 0.3, 'class3': 0.6}``.

        labels: List[str]
            A list of labels for ordering the axes of the confusion matrix.

        Returns
        -------
        accuracy: float
            The categorical accuracy of the classifier on the observation sequences.

        confusion: numpy.ndarray
            The confusion matrix representing the discrepancy between predicted and actual labels.
        """
        X, y = self._val.observation_sequences_and_labels(X, y)

        if labels is not None:
            self._val.list_of_strings(labels, desc='confusion matrix labels')

        # Classify each observation sequence and calculate confusion matrix
        predictions = self.predict(X, prior, return_scores=False)
        cm = confusion_matrix(y, predictions, labels=labels)

        return np.sum(np.diag(cm)) / np.sum(cm), cm

    def as_dict(self):
        """Serializes the :class:`HMMClassifier` object into a `dict`, ready to be stored in JSON format.

        .. note::
            Serializing a :class:`HMMClassifier` implicitly serializes the internal :class:`HMM` or :class:`GMMHMM` objects
            by calling :meth:`HMM.as_dict` or :meth:`GMMHMM.as_dict` and storing all of the model data in a single `dict`.

        Returns
        -------
        serialized: dict
            JSON-ready serialization of the :class:`HMMClassifier` object.

        See Also
        --------
        HMM.as_dict: The serialization function used for individual :class:`HMM` objects.
        GMMHMM.as_dict: The serialization function used for individual :class:`GMMHMM` objects.
        """

        try:
            self._models
        except AttributeError as e:
            raise AttributeError('The classifier needs to be fitted before it can be exported to a dict') from e

        return {'models': [model.as_dict() for model in self._models]}

    def save(self, path):
        """Converts the :class:`HMMClassifier` object into a `dict` and stores it in a JSON file.

        Parameters
        ----------
        path: str
            File path (with or without `.json` extension) to store the JSON-serialized :class:`HMMClassifier` object.

        See Also
        --------
        as_dict: Generates the `dict` that is stored in the JSON file.
        """

        data = self.as_dict()
        with open(path, 'w') as f:
            json.dump(data, f, indent=4)

    @classmethod
    def load(cls, path, random_state=None):
        """Deserializes either a `dict` or JSON serialized :class:`HMMClassifier` object.

        Parameters
        ----------
        path: str
            File path of the serialized JSON data generated by the :meth:`save` method.

        random_state: numpy.random.RandomState, int, optional
            A random state object or seed for reproducible randomness.

        Returns
        -------
        deserialized: :class:`HMMClassifier`
            The deserialized HMM classifier object.

        See Also
        --------
        save: Serializes a :class:`HMMClassifier` into a JSON file.
        as_dict: Generates a `dict` representation of the :class:`HMMClassifier`.
        """

        # Load the serialized HMM classifier as JSON
        with open(path, 'r') as f:
            data = json.load(f)

        clf = cls()
        clf._models = []

        for model in data['models']:
            # Retrieve the type of HMM
            if model['type'] == 'HMM':
                hmm = HMM
            elif model['type'] == 'GMMHMM':
                hmm = GMMHMM
            else:
                raise ValueError("Expected 'type' field to be either 'HMM' or 'GMMHMM'")

            # Deserialize the HMM and add it to the classifier
            clf._models.append(hmm.load(model, random_state=random_state))

        return clf


import numpy as np, pomegranate as pg, json
from .hmm import HMM

class GMMHMM(HMM):
    """A hidden Markov model representing an isolated temporal sequence class,
    with mixtures of multivariate Gaussian components representing state emission distributions.

    Parameters
    ----------
    label: str
        A label for the model, corresponding to the class being represented.

    n_states: int
        The number of states for the model.

    n_components: int
        The number of mixture components used in the emission distribution for each state.

    covariance: {'diagonal', 'full'}
        The covariance matrix type.

    topology: {'ergodic', 'left-right', 'strict-left-right'}
        The topology for the model.

    random_state: numpy.random.RandomState, int, optional
        A random state object or seed for reproducible randomness.

    Attributes
    ----------
    label: str
        The label for the model.

    n_states: int
        The number of states for the model.

    n_seqs: int
        The number of observation sequences use to train the model.

    initial: numpy.ndarray
        The initial state distribution of the model.

    transitions: numpy.ndarray
        The transition matrix of the model.
    """

    def __init__(self, label, n_states, n_components, covariance='diagonal', topology='left-right', random_state=None):
        super().__init__(label, n_states, topology, random_state)
        self._n_components = self._val.restricted_integer(
            n_components, lambda x: x > 1, desc='number of mixture components', expected='greater than one')
        self._covariance = self._val.one_of(covariance, ['diagonal', 'full'], desc='covariance matrix type')

    def fit(self, X, n_jobs=1):
        """Fits the HMM to observation sequences assumed to be labeled as the class that the model represents.

        Parameters
        ----------
        X: List[numpy.ndarray]
            Collection of multivariate observation sequences, each of shape :math:`(T \\times D)` where
            :math:`T` may vary per observation sequence.

        n_jobs: int
            | The number of jobs to run in parallel.
            | Setting this to -1 will use all available CPU cores.
        """
        X = self._val.observation_sequences(X)
        self._val.restricted_integer(n_jobs, lambda x: x == -1 or x > 0, 'number of jobs', '-1 or greater than zero')

        try:
            (self._initial, self._transitions)
        except AttributeError as e:
            raise AttributeError('Must specify initial state distribution and transitions before the HMM can be fitted') from e

        self._n_seqs = len(X)
        self._n_features = X[0].shape[1]

        # Create a mixture distribution of multivariate Gaussian emission components using combined samples for initial parameter estimation
        concat = np.concatenate(X)
        if self._covariance == 'diagonal':
            # Use diagonal covariance matrices
            dist = pg.GeneralMixtureModel(
                [pg.MultivariateGaussianDistribution(concat.mean(axis=0), concat.std(axis=0) * np.eye(self._n_features)) for _ in range(self._n_components)],
                self._random_state.dirichlet(np.ones(self._n_components)
            )
        )
        else:
            # Use full covariance matrices
            dist = pg.GeneralMixtureModel.from_samples(pg.MultivariateGaussianDistribution, self._n_components, concat)

        # Create the HMM object
        self._model = pg.HiddenMarkovModel.from_matrix(
            name=self._label,
            transition_probabilities=self._transitions,
            distributions=[dist.copy() for _ in range(self._n_states)],
            starts=self._initial
        )

        # Perform the Baum-Welch algorithm to fit the model to the observations
        self._model.fit(X, n_jobs=n_jobs)

        # Update the initial state distribution and transitions to reflect the updated parameters
        inner_tx = self._model.dense_transition_matrix()[:, :self._n_states]
        self._initial = inner_tx[self._n_states]
        self._transitions = inner_tx[:self._n_states]

    @property
    def n_components(self):
        return self._n_components

    @property
    def covariance(self):
        return self._covariance

    def as_dict(self):
        """Serializes the :class:`GMMHMM` object into a `dict`, ready to be stored in JSON format.

        Returns
        -------
        serialized: dict
            JSON-ready serialization of the :class:`GMMHMM` object.
        """

        try:
            self._model
        except AttributeError as e:
            raise AttributeError('The model needs to be fitted before it can be exported to a dict') from e

        model = self._model.to_json()

        if 'NaN' in model:
            raise ValueError('Encountered NaN value(s) in HMM parameters')
        else:
            return {
                'type': 'GMMHMM',
                'label': self._label,
                'n_states': self._n_states,
                'n_components': self._n_components,
                'covariance': self._covariance,
                'topology': self._topologies[self._topology.__class__],
                'model': {
                    'initial': self._initial.tolist(),
                    'transitions': self._transitions.tolist(),
                    'n_seqs': self._n_seqs,
                    'n_features': self._n_features,
                    'hmm': json.loads(model)
                }
            }

    @classmethod
    def load(cls, data, random_state=None):
        """Deserializes either a `dict` or JSON serialized :class:`GMMHMM` object.

        Parameters
        ----------
        data: str or dict
            - File path of the serialized JSON data generated by the :meth:`save` method.
            - `dict` representation of the :class:`GMMHMM`, generated by the :meth:`as_dict` method.

        random_state: numpy.random.RandomState, int, optional
            A random state object or seed for reproducible randomness.

        Returns
        -------
        deserialized: :class:`GMMHMM`
            The deserialized HMM object.

        See Also
        --------
        save: Serializes a :class:`GMMHMM` into a JSON file.
        as_dict: Generates a `dict` representation of the :class:`GMMHMM`.
        """

        # Load the serialized GMM-HMM data
        if isinstance(data, dict):
            pass
        elif isinstance(data, str):
            with open(data, 'r') as f:
                data = json.load(f)
        else:
            pass

        # Check that JSON is in the "correct" format
        if data['type'] == 'HMM':
            raise ValueError('You must use the HMM class to deserialize a stored HMM model')
        elif data['type'] == 'GMMHMM':
            pass
        else:
            raise ValueError("Attempted to deserialize an invalid model - expected 'type' field to be 'GMMHMM'")

        # Deserialize the data into a GMM-HMM object
        gmmhmm = cls(data['label'], data['n_states'], data['n_components'], data['covariance'], data['topology'], random_state=random_state)
        gmmhmm._initial = np.array(data['model']['initial'])
        gmmhmm._transitions = np.array(data['model']['transitions'])
        gmmhmm._n_seqs = data['model']['n_seqs']
        gmmhmm._n_features = data['model']['n_features']
        gmmhmm._model = pg.HiddenMarkovModel.from_json(json.dumps(data['model']['hmm']))

        return gmmhmm